#include "stdafx.h"
#include "Disassembly.h"
#include "MachineState.h"
#include "MainForm.h"

using namespace Simulator;
using namespace System::Windows::Forms;
using namespace System::IO;

Disassembly::Disassembly(MachineState^ machineState)
{	
	m_machineState = machineState;

	m_opcodeLookup[0xFF] = InstructionData("HLT", 1, 0);
	m_opcodeLookup[0xEA] = InstructionData("NOP", 1, 0);
	m_opcodeLookup[0xEF] = InstructionData("SPX", 1, 0);
	m_opcodeLookup[0x4C] = InstructionData("JMP", 3, 0);
	m_opcodeLookup[0xA5] = InstructionData("LDA", 2, 0);
	m_opcodeLookup[0xAD] = InstructionData("LDA", 3, 0);
	m_opcodeLookup[0xA9] = InstructionData("LDA", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xAF] = InstructionData("LDA", 4, 0);
	m_opcodeLookup[0xA6] = InstructionData("LDX", 2, 0);
	m_opcodeLookup[0xAE] = InstructionData("LDX", 3, 0);
	m_opcodeLookup[0xA2] = InstructionData("LDX", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xBE] = InstructionData("LDX", 3, InstructionData::kYIndexed);
	m_opcodeLookup[0xA4] = InstructionData("LDY", 2, 0);
	m_opcodeLookup[0xAC] = InstructionData("LDY", 3, 0);
	m_opcodeLookup[0xA0] = InstructionData("LDY", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xAA] = InstructionData("TAX", 1, 0);
	m_opcodeLookup[0x8A] = InstructionData("TXA", 1, 0);
	m_opcodeLookup[0xA8] = InstructionData("TAY", 1, 0);	
	m_opcodeLookup[0x98] = InstructionData("TYA", 1, 0);
	m_opcodeLookup[0x9B] = InstructionData("TXY", 1, 0);	
	m_opcodeLookup[0xBB] = InstructionData("TYX", 1, 0);
	m_opcodeLookup[0x48] = InstructionData("PHA", 1, 0);
	m_opcodeLookup[0x68] = InstructionData("PLA", 1, 0);
	m_opcodeLookup[0xDA] = InstructionData("PHX", 1, 0);
	m_opcodeLookup[0xFA] = InstructionData("PLX", 1, 0);
	m_opcodeLookup[0x5A] = InstructionData("PHY", 1, 0);
	m_opcodeLookup[0x7A] = InstructionData("PLY", 1, 0);
	m_opcodeLookup[0x80] = InstructionData("BRA", 2, InstructionData::kBranch);
	m_opcodeLookup[0xF0] = InstructionData("BEQ", 2, InstructionData::kBranch);
	m_opcodeLookup[0xD0] = InstructionData("BNE", 2, InstructionData::kBranch);
	m_opcodeLookup[0xB0] = InstructionData("BCS", 2, InstructionData::kBranch);
  	m_opcodeLookup[0x90] = InstructionData("BCC", 2, InstructionData::kBranch);
	m_opcodeLookup[0x70] = InstructionData("BVS", 2, InstructionData::kBranch);
  	m_opcodeLookup[0x50] = InstructionData("BVC", 2, InstructionData::kBranch);
	m_opcodeLookup[0x30] = InstructionData("BMI", 2, InstructionData::kBranch);
   	m_opcodeLookup[0x10] = InstructionData("BPL", 2, InstructionData::kBranch);
	m_opcodeLookup[0xE8] = InstructionData("INX", 1, 0);
	m_opcodeLookup[0xCA] = InstructionData("DEX", 1, 0);  
	m_opcodeLookup[0xC8] = InstructionData("INY", 1, 0);
	m_opcodeLookup[0x88] = InstructionData("DEY", 1, 0);
	m_opcodeLookup[0x3A] = InstructionData("DEC A", 1, 0);
	m_opcodeLookup[0xC6] = InstructionData("DEC", 2, 0);
	m_opcodeLookup[0xCE] = InstructionData("DEC", 3, 0);
	m_opcodeLookup[0x1A] = InstructionData("INC A", 1, 0);
	m_opcodeLookup[0xE6] = InstructionData("INC", 2, 0);
    m_opcodeLookup[0xEE] = InstructionData("INC", 3, 0);
	m_opcodeLookup[0x85] = InstructionData("STA", 2, 0);
	m_opcodeLookup[0x8D] = InstructionData("STA", 3, 0);
	m_opcodeLookup[0x8F] = InstructionData("STA", 4, 0);
	m_opcodeLookup[0x86] = InstructionData("STX", 2, 0);
	m_opcodeLookup[0x8E] = InstructionData("STX", 3, 0);
	m_opcodeLookup[0x84] = InstructionData("STY", 2, 0);
	m_opcodeLookup[0x8C] = InstructionData("STY", 3, 0);
	m_opcodeLookup[0x20] = InstructionData("JSR", 3, 0);
	m_opcodeLookup[0x60] = InstructionData("RTS", 1, 0);
	m_opcodeLookup[0x0A] = InstructionData("ASL A", 1, 0);
	m_opcodeLookup[0x06] = InstructionData("ASL", 2, 0);
	m_opcodeLookup[0x0E] = InstructionData("ASL", 3, 0);
	m_opcodeLookup[0x4A] = InstructionData("LSR A", 1, 0);
	m_opcodeLookup[0x46] = InstructionData("LSR", 2, 0);
	m_opcodeLookup[0x4E] = InstructionData("LSR", 3, 0);
	m_opcodeLookup[0x2A] = InstructionData("ROL A", 1, 0);
	m_opcodeLookup[0x26] = InstructionData("ROL", 2, 0);
	m_opcodeLookup[0x2E] = InstructionData("ROL", 3, 0);
	m_opcodeLookup[0x6A] = InstructionData("ROR A", 1, 0);
	m_opcodeLookup[0x66] = InstructionData("ROR", 2, 0);
	m_opcodeLookup[0x6E] = InstructionData("ROR", 3, 0);
	m_opcodeLookup[0x38] = InstructionData("SEC", 1, 0);
	m_opcodeLookup[0x18] = InstructionData("CLC", 1, 0);	
	m_opcodeLookup[0xB8] = InstructionData("CLV", 1, 0);
	m_opcodeLookup[0x08] = InstructionData("PHP", 1, 0);
	m_opcodeLookup[0x28] = InstructionData("PLP", 1, 0);
	m_opcodeLookup[0x8B] = InstructionData("PHB", 1, 0);
	m_opcodeLookup[0xAB] = InstructionData("PLB", 1, 0);
	m_opcodeLookup[0x00] = InstructionData("BRK", 1, 0);
	m_opcodeLookup[0x40] = InstructionData("RTI", 1, 0);
	m_opcodeLookup[0xC9] = InstructionData("CMP", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xC5] = InstructionData("CMP", 2, 0);
	m_opcodeLookup[0xCD] = InstructionData("CMP", 3, 0);
	m_opcodeLookup[0xE0] = InstructionData("CPX", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xE4] = InstructionData("CPX", 2, 0);
	m_opcodeLookup[0xEC] = InstructionData("CPX", 3, 0);
	m_opcodeLookup[0xC0] = InstructionData("CPY", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xC4] = InstructionData("CPY", 2, 0);
	m_opcodeLookup[0xCC] = InstructionData("CPY", 3, 0);
	m_opcodeLookup[0x58] = InstructionData("CLI", 1, 0);
	m_opcodeLookup[0x78] = InstructionData("SEI", 1, 0);
	m_opcodeLookup[0x09] = InstructionData("ORA", 2, InstructionData::kImmediate);
	m_opcodeLookup[0x05] = InstructionData("ORA", 2, 0);
	m_opcodeLookup[0x0D] = InstructionData("ORA", 3, 0);
	m_opcodeLookup[0x29] = InstructionData("AND", 2, InstructionData::kImmediate);
	m_opcodeLookup[0x25] = InstructionData("AND", 2, 0);
	m_opcodeLookup[0x2D] = InstructionData("AND", 3, 0);
	m_opcodeLookup[0x49] = InstructionData("EOR", 2, InstructionData::kImmediate);
	m_opcodeLookup[0x45] = InstructionData("EOR", 2, 0);
	m_opcodeLookup[0x4D] = InstructionData("EOR", 3, 0);
	m_opcodeLookup[0x69] = InstructionData("ADC", 2, InstructionData::kImmediate);
	m_opcodeLookup[0x65] = InstructionData("ADC", 2, 0);
	m_opcodeLookup[0x6D] = InstructionData("ADC", 3, 0);
	m_opcodeLookup[0xE9] = InstructionData("SBC", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xE5] = InstructionData("SBC", 2, 0);
	m_opcodeLookup[0xED] = InstructionData("SBC", 3, 0);
	m_opcodeLookup[0x89] = InstructionData("BIT", 2, InstructionData::kImmediate);
	m_opcodeLookup[0x24] = InstructionData("BIT", 2, 0);
	m_opcodeLookup[0x2C] = InstructionData("BIT", 3, 0);
	m_opcodeLookup[0xDD] = InstructionData("CMP", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0xB5] = InstructionData("LDA", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0xBD] = InstructionData("LDA", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0xB9] = InstructionData("LDA", 3, InstructionData::kYIndexed);
	m_opcodeLookup[0xB4] = InstructionData("LDY", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0xBC] = InstructionData("LDY", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x95] = InstructionData("STA", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0x9D] = InstructionData("STA", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x99] = InstructionData("STA", 3, InstructionData::kYIndexed);
	m_opcodeLookup[0x94] = InstructionData("STY", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0xDE] = InstructionData("DEC", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0xF6] = InstructionData("INC", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0xFE] = InstructionData("INC", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x16] = InstructionData("ASL", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0x1E] = InstructionData("ASL", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x56] = InstructionData("LSR", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0x5E] = InstructionData("LSR", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x3E] = InstructionData("ROL", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x76] = InstructionData("ROR", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0x7E] = InstructionData("ROR", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x1D] = InstructionData("ORA", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x3D] = InstructionData("AND", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x5D] = InstructionData("EOR", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x7D] = InstructionData("ADC", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0x79] = InstructionData("ADC", 3, InstructionData::kYIndexed);
	m_opcodeLookup[0xF5] = InstructionData("SBC", 2, InstructionData::kXIndexed);
	m_opcodeLookup[0xFD] = InstructionData("SBC", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0xF9] = InstructionData("SBC", 3, InstructionData::kYIndexed);
	m_opcodeLookup[0x6C] = InstructionData("JMP", 3, InstructionData::kIndirect);
	m_opcodeLookup[0xEB] = InstructionData("BANK", 2, InstructionData::kImmediate);
	m_opcodeLookup[0xFB] = InstructionData("BANK", 1, 0);
	m_opcodeLookup[0x1B] = InstructionData("TRS", 1, 0);
	m_opcodeLookup[0x0B] = InstructionData("TSR", 1, 0);
	m_opcodeLookup[0x9A] = InstructionData("TXS", 1, 0);
	m_opcodeLookup[0xBA] = InstructionData("TSX", 1, 0);
	m_opcodeLookup[0x9C] = InstructionData("CLR", 3, 0);
	m_opcodeLookup[0x9E] = InstructionData("CLR", 3, InstructionData::kXIndexed);
	m_opcodeLookup[0xB2] = InstructionData("LDA", 3, InstructionData::kIndirect | InstructionData::kXIndexed);
	m_opcodeLookup[0x92] = InstructionData("STA", 3, InstructionData::kIndirect | InstructionData::kXIndexed);
	m_opcodeLookup[0xB1] = InstructionData("LDA", 2, InstructionData::kIndirect | InstructionData::kYIndexed);
	m_opcodeLookup[0x91] = InstructionData("STA", 2, InstructionData::kIndirect | InstructionData::kYIndexed);
	m_opcodeLookup[0xD1] = InstructionData("CMP", 2, InstructionData::kIndirect | InstructionData::kYIndexed);
	m_opcodeLookup[0xF1] = InstructionData("SBC", 2, InstructionData::kIndirect | InstructionData::kYIndexed);
	m_opcodeLookup[0xA3] = InstructionData("LDA", 2, InstructionData::kStackRelative);
	m_opcodeLookup[0x83] = InstructionData("STA", 2, InstructionData::kStackRelative);
	m_opcodeLookup[0xB3] = InstructionData("LDA", 2, InstructionData::kStackRelative | InstructionData::kIndirect | InstructionData::kXIndexed);
	m_opcodeLookup[0x93] = InstructionData("STA", 2, InstructionData::kStackRelative | InstructionData::kIndirect | InstructionData::kXIndexed);
	m_opcodeLookup[0x5C] = InstructionData("JML", 4, 0);
	m_opcodeLookup[0x22] = InstructionData("JSL", 4, 0);
	m_opcodeLookup[0x6B] = InstructionData("RTL", 1, 0);
}

int Disassembly::GetOpcodeSize(unsigned char opcode)
{
	if (m_opcodeLookup.ContainsKey(opcode))
	{
		return m_opcodeLookup[opcode].Length;
	}
	else
	{
		return 0;
	}
}

String^ Disassembly::GetOpcodeName(unsigned char opcode)
{
	if (m_opcodeLookup.ContainsKey(opcode))
	{
		return m_opcodeLookup[opcode].Mnemo;
	}
	else
	{
		return "???";
	}
}

String^ Disassembly::GetOpcodeDescription(unsigned char opcode)
{

	if (m_opcodeLookup.ContainsKey(opcode))
	{
		InstructionData data = m_opcodeLookup[opcode];
		String^ result = data.Mnemo;
		if (data.FlagData & InstructionData::kImmediate)
		{
			result += " #";
		}
		else if (!(data.FlagData & InstructionData::kBranch))
		{
			if (data.Length == 2 && !(data.FlagData & InstructionData::kStackRelative))
			{
				result += " zp";
			}

			if (data.FlagData & InstructionData::kIndirect)
			{
				result += " ind";
			}
			else
			{
				result += " abs";	
			}

			if (data.FlagData & InstructionData::kXIndexed)
			{
				result += ",X";
			}
			else if (data.FlagData & InstructionData::kYIndexed)
			{
				result += ",Y";
			}
		}

		return result;
	}
	else
	{
		return "???";
	}
}


void Disassembly::FillDisassemblyListView(ListView^ listView, unsigned long startAddress, unsigned long pcAddress, unsigned long stackFrameAddress, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, Dictionary<String^, unsigned long>^ sourceStartAddress, bool ignoreSource)
{
	if (startAddress == 0xFFFFFFFF)
	{
		startAddress = 0;
	}

	const int linesToShow = 300;
	int currentInstructionIndex = -1;
	int stackFrameIndex = -1;
	int instructionLength;
	unsigned long nextAddress = startAddress;
	unsigned long setBreakpointAddr = 0;
	bool foundStackFrameAddr = false;

	listView->BeginUpdate();
	listView->Items->Clear();

	bool foundPC = false;

	// look for a source mapping
	bool foundSource = false;
	FileAndLine fileAndLine;
	if (!ignoreSource &&
		(addressToSource->TryGetValue(stackFrameAddress, fileAndLine) ||
		 addressToSource->TryGetValue(stackFrameAddress-1, fileAndLine) ||
		 addressToSource->TryGetValue(stackFrameAddress-2, fileAndLine)))
	{
		try
		{
			StreamReader^ sr = gcnew StreamReader(fileAndLine.filename);

			try
			{
				String^ line;
			 
				// Read and display lines from the file until the end of 
				// the file is reached.
				int lineNumber = 1;
				nextAddress = sourceStartAddress[fileAndLine.filename];
				while ( (line = sr->ReadLine()) != nullptr )
				{
					line = line->Replace("\t", "    ");
					ListViewItem^ lineItem = gcnew ListViewItem(line);
					lineItem->Tag = nextAddress; 

					listView->Items->Add(lineItem);

					if (lineNumber == fileAndLine.lineNumber && stackFrameAddress == pcAddress && !foundPC)
					{
						if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
						{
							lineItem->ImageKey = "breakpoint with arrow";
							setBreakpointAddr = nextAddress;
						}
						else
						{
							lineItem->ImageKey = "arrow";
						}
						currentInstructionIndex = lineItem->Index;
						foundPC = true;

						if (stackFrameAddress == nextAddress)
						{
							foundStackFrameAddr = true;
						}
					}
					else if (stackFrameAddress != pcAddress && stackFrameAddress == nextAddress && !foundStackFrameAddr)
					{
						if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
						{
							lineItem->ImageKey = "breakpoint with stackframe";
							setBreakpointAddr = nextAddress;
						}
						else
						{
							lineItem->ImageKey = "stackframe";
						}
						stackFrameIndex = lineItem->Index;
						foundStackFrameAddr = true;
					}
					else if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
					{
						lineItem->ImageKey = "breakpoint";
						setBreakpointAddr = nextAddress;
					}

					if (!sourceBreakpoints->ContainsKey(nextAddress))
					{	
						setBreakpointAddr = 0;
					}

					lineNumber++;

					int searchCount = 0;
					unsigned long testAddress = nextAddress;
					while (searchCount++ < 32) // skip up to 32 bytes
					{
						testAddress++;
						FileAndLine newFileAndLine;
						if (addressToSource->TryGetValue(testAddress, newFileAndLine))
						{
							if (newFileAndLine.lineNumber == lineNumber && 
								newFileAndLine.filename == fileAndLine.filename)
							{
								nextAddress = testAddress;
								break;
							}			
						}

					}

					if (searchCount > 32)
					{
						// didn't find it by looking up the key, so check all the values
						for each (KeyValuePair<unsigned long,FileAndLine> kvp in addressToSource)
						{
							if (kvp.Value.lineNumber == lineNumber && 
								kvp.Value.filename == fileAndLine.filename)
							{
								nextAddress = kvp.Key;
								break;
							}
						}
					}
				}

				foundSource = true;
			}
			finally
			{
				if ( sr )
				delete (IDisposable^)sr;	
			}
		}
		catch (Exception^ )
		{
		}
	}
	
	if (!foundSource)
	{
		// disassembly
		for (int i=0; i<linesToShow; i++)
		{
			ListViewItem^ lineItem = gcnew ListViewItem(String::Format("{0:X4} ", nextAddress));
			lineItem->Tag = nextAddress;

			FileAndLine fileAndLine;
			if (addressToSource->TryGetValue(nextAddress, fileAndLine))
			{
				lineItem->Text += fileAndLine.sourceLabel;
				for (int i=fileAndLine.sourceLabel->Length; i<21; i++)
				{
					lineItem->Text += " ";
				}
			}
			else
			{
				lineItem->Text += "                     ";
			}

			unsigned char opcode = m_machineState->Peek(nextAddress);
			if (m_opcodeLookup.ContainsKey(opcode))
			{
				InstructionData% data = m_opcodeLookup[opcode];
				if (data.FlagData & InstructionData::kBranch)
				{
					unsigned char branchDistance = m_machineState->Peek(nextAddress+1);
					unsigned long destination = branchDistance > 0x80 ? nextAddress + 2 - (256-branchDistance) : nextAddress + branchDistance + 2;
					
					// with the current microcode, destination bank for branches is same as source bank
					destination &= 0x0000FFFF;
					//destination |= ((nextAddress + 2) & 0x00FF0000);

					lineItem->Text += String::Format("{0} ${1:X4}", data.Mnemo, destination);
				}
				else if (data.Length == 1)
				{
					lineItem->Text += data.Mnemo;
				}
				else if (data.Length == 2)
				{
					if (data.FlagData & InstructionData::kStackRelative)
					{
						if (data.FlagData & InstructionData::kIndirect)
						{
							lineItem->Text += String::Format("{0} ({1},S)\r\n", data.Mnemo, m_machineState->Peek(nextAddress+1));
						}
						else
						{
							lineItem->Text += String::Format("{0} {1},S\r\n", data.Mnemo, m_machineState->Peek(nextAddress+1));
						}

						if (data.FlagData & InstructionData::kXIndexed)
						{
							lineItem->Text += ",X";
						}
					}
					else if (data.FlagData & InstructionData::kImmediate)
					{
						lineItem->Text += String::Format("{0} #${1:X2}\r\n", data.Mnemo, m_machineState->Peek(nextAddress+1));
					}
					else
					{
						// zero page
						unsigned int address = m_machineState->Peek(nextAddress+1);

						if (data.FlagData & InstructionData::kIndirect)
						{
							lineItem->Text += String::Format("{0} (${1:X2})", data.Mnemo, address);
						}
						else
						{
							lineItem->Text += String::Format("{0} ${1:X2}", data.Mnemo, address);		
						}

						if (data.FlagData & InstructionData::kXIndexed)
						{
							lineItem->Text += ",X";
						}
						else if (data.FlagData & InstructionData::kYIndexed)
						{
							lineItem->Text += ",Y";
						}
					}
				}
				else if (data.Length == 3)
				{
					unsigned int address = m_machineState->Peek(nextAddress+1) + 256*m_machineState->Peek(nextAddress+2);

					if (data.FlagData & InstructionData::kIndirect)
					{
						lineItem->Text += String::Format("{0} (${1:X4})", data.Mnemo, address);
					}
					else
					{
						lineItem->Text += String::Format("{0} ${1:X4}", data.Mnemo, address);		
					}

					if (data.FlagData & InstructionData::kXIndexed)
					{
						lineItem->Text += ",X";
					}
					else if (data.FlagData & InstructionData::kYIndexed)
					{
						lineItem->Text += ",Y";
					}
				}

				else if (data.Length == 4)
				{
					unsigned int address = m_machineState->Peek(nextAddress+1) + 256*m_machineState->Peek(nextAddress+2) + 65536*m_machineState->Peek(nextAddress+3);

					if (data.FlagData & InstructionData::kIndirect)
					{
						lineItem->Text += String::Format("{0} (${1:X6})", data.Mnemo, address);
					}
					else
					{
						lineItem->Text += String::Format("{0} ${1:X6}", data.Mnemo, address);		
					}

					if (data.FlagData & InstructionData::kXIndexed)
					{
						lineItem->Text += ",X";
					}
				}

				instructionLength = data.Length;
			}
			else
			{
				lineItem->Text += "???";
				instructionLength = 1;
			}

			listView->Items->Add(lineItem);

			if (nextAddress == pcAddress && !foundPC)
			{
				if (sourceBreakpoints->ContainsKey(nextAddress))
				{
					lineItem->ImageKey = "breakpoint with arrow";
				}
				else
				{
					lineItem->ImageKey = "arrow";
				}
				currentInstructionIndex = lineItem->Index;
				foundPC = true;
				
				if (stackFrameAddress == nextAddress)
				{
					foundStackFrameAddr = true;
				}
			}
			else if (stackFrameAddress != pcAddress && stackFrameAddress == nextAddress && !foundStackFrameAddr)
			{
				if (sourceBreakpoints->ContainsKey(nextAddress) && setBreakpointAddr != nextAddress)
				{
					lineItem->ImageKey = "breakpoint with stackframe";
					setBreakpointAddr = nextAddress;
				}
				else
				{
					lineItem->ImageKey = "stackframe";
				}
				stackFrameIndex = lineItem->Index;
				foundStackFrameAddr = true;
			}
			else if (sourceBreakpoints->ContainsKey(nextAddress))
			{
				lineItem->ImageKey = "breakpoint";
			}

			nextAddress += instructionLength;
		}
	}

	listView->EndUpdate();

	if (stackFrameIndex >= 0)
	{
		listView->EnsureVisible(stackFrameIndex);
	}
	else if (currentInstructionIndex >= 0)
	{
		listView->EnsureVisible(currentInstructionIndex);
	}
}

void Disassembly::UpdateDisassemblyListView(ListView^ listView, unsigned long pcAddress, unsigned long stackFrameAddress, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, bool scrollToCurrentInstruction)
{
	listView->BeginUpdate();

	bool foundPC = false;
	unsigned long setBreakpointAddr = 0;
	bool foundStackFrameAddr = false;
	int currentInstructionIndex = -1;
	int stackFrameIndex = -1;

	for each (ListViewItem^ item in listView->Items)
	{
		unsigned long address = safe_cast<unsigned long>(item->Tag);
		if (address == pcAddress && !foundPC)
		{
			if (sourceBreakpoints->ContainsKey(address) && setBreakpointAddr != address)
			{
				item->ImageKey = "breakpoint with arrow";
				setBreakpointAddr = address;
			}
			else
			{
				item->ImageKey = "arrow";
			}
			
			if (stackFrameAddress == address)
			{
				foundStackFrameAddr = true;
			}

			currentInstructionIndex = item->Index;

			foundPC = true;

		}
		else if (stackFrameAddress != pcAddress && stackFrameAddress == address && !foundStackFrameAddr)
		{
			if (sourceBreakpoints->ContainsKey(address) && setBreakpointAddr != address)
			{
				item->ImageKey = "breakpoint with stackframe";
				setBreakpointAddr = address;
			}
			else
			{
				item->ImageKey = "stackframe";
			}
			stackFrameIndex = item->Index;
			foundStackFrameAddr = true;
		}
		else if (sourceBreakpoints->ContainsKey(address) && setBreakpointAddr != address)
		{
			item->ImageKey = "breakpoint";
			setBreakpointAddr = address;
		}
		else
		{
			item->ImageKey = nullptr;
		}

		if (!sourceBreakpoints->ContainsKey(address))
		{
			setBreakpointAddr = 0;
		}
	}

	if (scrollToCurrentInstruction)
	{
		if (stackFrameIndex >= 0)
		{
			listView->EnsureVisible(stackFrameIndex);
		}
		else if (currentInstructionIndex >= 0)
		{
			listView->EnsureVisible(currentInstructionIndex);
		}
	}

	listView->EndUpdate();
}
