#pragma once

using namespace System::Windows::Forms;
using namespace System::Collections::Generic;

namespace Simulator {

	ref class MachineState;
	value class FileAndLine;

	using namespace System;

	public value class InstructionData
	{
	public:
		static const unsigned int kBranch = 1;
		static const unsigned int kXIndexed = 2;
		static const unsigned int kIndirect = 4;
		static const unsigned int kStackRelative = 8;
		static const unsigned int kImmediate = 16;
		static const unsigned int kYIndexed = 32;

		InstructionData::InstructionData(String^ mnemo, int length, unsigned long flagData)
		{
			Mnemo = mnemo;
			Length = length;
			FlagData = flagData;
		}

		property String^ Mnemo;
		property int Length;
		property unsigned long FlagData;
	};

	public ref class Disassembly 
	{
	public:
		Disassembly(MachineState^ machineState);

		void FillDisassemblyListView(ListView^ listView, unsigned long startAddress, unsigned long pcAddress, unsigned long stackFrameAddress, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, Dictionary<String^, unsigned long>^ sourceStartAddress, bool ignoreSource);
		void UpdateDisassemblyListView(ListView^ listView, unsigned long pcAddress, unsigned long stackFrameAddress, Dictionary<unsigned long, bool>^ sourceBreakpoints, Dictionary<unsigned long, FileAndLine>^ addressToSource, bool scrollToCurrentInstruction);
		String^ GetOpcodeName(unsigned char opcode);
		String^ GetOpcodeDescription(unsigned char opcode);
		int GetOpcodeSize(unsigned char opcode);

	protected:
		~Disassembly() {}
		Dictionary<unsigned char, InstructionData> m_opcodeLookup;
		MachineState^ m_machineState;
	};
}

