#include "stdafx.h"
#include "MemorySystem.h"

using namespace Simulator;
using namespace System::IO;
using namespace System::Globalization;
using namespace System::Windows::Forms;

MemorySystem::MemorySystem(String^ romPath, unsigned long loadAddress) :
	m_ram(gcnew array<unsigned char>(s_ramSize)),
	m_lcdPos(0),
	m_lcdText("                                                                                ")
{
	try 
	{
		FileStream^  fs = nullptr; 	
		BinaryReader^ br = nullptr; 		

		fs = File::OpenRead(romPath); 	
		br = gcnew BinaryReader(fs); 
		unsigned long addr = loadAddress;

		bool done = false;
		while (!done)
		{
			try
			{
				m_ram[addr] = br->ReadByte();
				addr++;
			}
			catch (System::IO::EndOfStreamException^)
			{
				done = true;
			}
		}

		br->Close();
		fs->Close();

	}
	catch (Exception^)
	{
		MessageBox::Show(String::Format("Error opening program file {0}", romPath));
	}
}

// Perform a load of a byte from ROM, RAM, or HW device. May change state of HW devices.
unsigned char MemorySystem::Load(unsigned long address)
{
	if (address >= 0x010100 && address < 0x010200)
	{
		//MessageBox::Show("accessed 6502 stack");
	}

	if (address == 0x3F2D)
	{
		// RTC_ADR_FLAGS should always look as if the periodic timer has expired
		return 0x04;
	}
	else if (address == 0x3F12) 
	{
		// keyboard data memory-mapped location	
		return m_ram[address];
	}
	else if (address == 0x3F13) 
	{
		// keyboard control memory-mapped location	
		unsigned char result = m_ram[address];
		m_ram[address] = 1; // clear IRQ bit after read
		return result;
	}
	else if (address < s_ramSize)
	{
		return m_ram[address];
	}
	else
	{
		throw gcnew System::ApplicationException("Address reference outside intalled memory");
	}
}

// "Peek" at a byte of data without changing any machine state
unsigned char MemorySystem::Peek(unsigned long address)
{ 
	if (address == 0x3F12) 
	{
		// keyboard data memory-mapped location	
		return m_ram[address];
	}
	else if (address == 0x3F13) 
	{
		// keyboard control memory-mapped location	
		return m_ram[address];
	}
	else if (address < s_ramSize)
	{
		return m_ram[address];
	}
	else
	{
		throw gcnew System::ApplicationException("Address reference outside intalled memory");
	}
}

// Store a byte in RAM or HW device
void MemorySystem::Store(unsigned long address, unsigned char value)
{
	if (address >= 0x010100 && address < 0x010200)
	{
		//MessageBox::Show("accessed 6502 stack");
	}

	if (address < 0x3F00)
	{
		//MessageBox::Show("wrote to ROM");
	}
	else if (address == 0x3F01)
	{
		// LCD control
		if (value == 0x01)
		{
			// clear/home
			m_lcdText = "                                                                                ";
			m_lcdPos = 0;
		}
		else if (value == 0x02)
		{
			// home
			m_lcdPos = 0;
		}
		else if (value & 0x80)
		{
			// cursor pos
			m_lcdPos = value & 0x7F;
		}
	}
	else if (address == 0x3F00)
	{
		// LCD data
		int stringPos = 0;
		if (m_lcdPos < 24)
		{
			stringPos = m_lcdPos;
		}
		else if (m_lcdPos >= 64 && m_lcdPos < 88)
		{
			stringPos = m_lcdPos - 40;
		}
		else
		{
			stringPos = -1;
		}

		String^ preText = gcnew String("");
		String^ postText = gcnew String("");
		if (stringPos > 0)
		{
			preText = m_lcdText->Substring(0, stringPos);
		}
		if (stringPos < 47)
		{
			postText = m_lcdText->Substring(stringPos + 1);
		}

		m_lcdText = preText + Char::ConvertFromUtf32(value) + postText;	

		m_lcdPos++;

		if (m_lcdPos == 24)
		{
			m_lcdPos = 64;
		}
		else if (m_lcdPos == 88)
		{
			m_lcdPos = 0;
		}
	}
	else if (address >= 0x3F00 && address < 0x4000)
	{
		// other memory-mapped hardware: ignore
	}
	else if (address < s_romSize)
	{
		MessageBox::Show("wrote to ROM");
	}
	else if (address < s_ramSize)
	{
		m_ram[address] = value;
	}
	else
	{
		throw gcnew System::ApplicationException("Address reference outside intalled memory");
	}
}